/* Copyright (c) 2016 Nordic Semiconductor. All Rights Reserved.
 *
 * The information contained herein is property of Nordic Semiconductor ASA.
 * Terms and conditions of usage are described in detail in NORDIC
 * SEMICONDUCTOR STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information. NO
 * WARRANTY of ANY KIND is provided. This heading must NOT be removed from
 * the file.
 *
 */

/** @file
 *
 * @brief MQTT Client Implementation TLS layer.
 *
 * This file contains the source code for MQTT Protocol TLS layer for a nRF device.
 * The implementation is limited to MQTT Client role only.
 */


#include "mem_manager.h"
#include "mqtt_internal.h"
#include "mqtt_rx.h"


uint32_t mqtt_client_tls_output_handler(nrf_tls_instance_t const * p_instance,
                                        uint8_t            const * p_data,
                                        uint32_t                   datalen)
{
    NULL_PARAM_CHECK(p_instance);

    uint32_t          err_code = NRF_ERROR_INTERNAL;
    mqtt_client_t   * p_client = (mqtt_client_t *)p_instance->transport_id;

    MQTT_MUTEX_LOCK ();

    MQTT_TRC("[MQTT]: << mqtt_client_tls_output_handler, client %p\r\n", p_client);

    if (p_client != NULL)
    {
        err_code = mqtt_client_tcp_write(p_client, p_data, datalen);
    }

    MQTT_TRC("[MQTT]: >> mqtt_client_tls_output_handler, client %p, result 0x%08x\r\n",
             p_client, err_code);

    MQTT_MUTEX_UNLOCK ();

    return err_code;
}


uint32_t mqtt_client_tls_connect(mqtt_client_t * p_client)
{
    const nrf_tls_options_t tls_option =
    {
        .output_fn      = mqtt_client_tls_output_handler,
        .transport_type = NRF_TLS_TYPE_STREAM,
        .role           = NRF_TLS_ROLE_CLIENT,
        .p_key_settings = p_client->p_security_settings
    };

    connect_request_encode(p_client,
                           &p_client->p_pending_packet,
                           &p_client->pending_packetlen);

    p_client->tls_instance.transport_id = (uint32_t)p_client;

    MQTT_MUTEX_UNLOCK ();

    uint32_t err_code = nrf_tls_alloc(&p_client->tls_instance, &tls_option);

    MQTT_MUTEX_LOCK ();

    return err_code;
}


uint32_t mqtt_client_tls_write(mqtt_client_t * p_client,
                               uint8_t const * p_data,
                               uint32_t        datalen)
{
    MQTT_MUTEX_UNLOCK ();

    //MQTT_TRC("[MQTT]: << mqtt_client_tls_write\r\n");
    uint32_t err_code = nrf_tls_write(&p_client->tls_instance, p_data, &datalen);

    MQTT_MUTEX_LOCK ();

    return err_code;
}


uint32_t mqtt_client_tls_read(mqtt_client_t * p_client, uint8_t * p_data, uint32_t datalen)
{
    //MQTT_TRC("[MQTT]: << mqtt_client_tls_read\r\n");

    uint32_t err = nrf_tls_input(&p_client->tls_instance, p_data, datalen);

    if ((err == NRF_SUCCESS) && (p_client->p_pending_packet == NULL))
    {
        uint32_t   rx_datalen = 1024;
        uint8_t  * p_data     = nrf_malloc(1024);

        if (p_data != NULL)
        {
            MQTT_MUTEX_UNLOCK ();

            err = nrf_tls_read(&p_client->tls_instance,
                               p_data,
                               &rx_datalen);

            MQTT_MUTEX_LOCK ();

            if ((err == NRF_SUCCESS) && (rx_datalen > 0))
            {
                 err = mqtt_handle_rx_data(p_client, p_data, rx_datalen);
            }

            nrf_free(p_data);
        }
    }

    return err;
}


uint32_t mqtt_client_tls_disconnect(mqtt_client_t * p_client)
{
    return mqtt_client_tcp_disconnect(p_client);
}
